#include "stm32l4xx_hal.h"
#include "globals.h"
#include "serial.h"

extern UART_HandleTypeDef huart1;
extern UART_HandleTypeDef huart2;
extern DMA_HandleTypeDef hdma_usart1_rx;
extern DMA_HandleTypeDef hdma_usart2_rx;

// 115200 -> 11K char/s -> 180 char/image
#define SERIAL_BUFFER_RX_SIZE 256

struct dmasinfos
{
	DMA_HandleTypeDef *hdma;
	uint32_t start;
	uint32_t end;
	uint32_t rd_pointer;
	int decr;
};

uint8_t serialBufferUSB[SERIAL_BUFFER_RX_SIZE];
uint8_t serialBufferCMOS[SERIAL_BUFFER_RX_SIZE];

struct dmasinfos serialdmaUSB, serialdmaCMOS;

// private
void initSerialDmasStructs(struct dmasinfos *dmainfo, DMA_HandleTypeDef *hdma, uint8_t *p, int size)
{
	dmainfo->hdma = hdma;
	dmainfo->start = dmainfo->rd_pointer =  (uint32_t) p;
  dmainfo->end = dmainfo->start+size;
	dmainfo->decr = SERIAL_BUFFER_RX_SIZE;
}

// public
void startSerial(void)
{
	initSerialDmasStructs(&serialdmaUSB, &hdma_usart2_rx, serialBufferUSB, SERIAL_BUFFER_RX_SIZE);
	initSerialDmasStructs(&serialdmaCMOS, &hdma_usart1_rx, serialBufferCMOS, SERIAL_BUFFER_RX_SIZE);

	HAL_UART_Receive_DMA(&huart2,serialBufferUSB,SERIAL_BUFFER_RX_SIZE);
	HAL_UART_Receive_DMA(&huart1,serialBufferCMOS,SERIAL_BUFFER_RX_SIZE);
}

// private
// Assumption : getOneC called faster than the serial buffer gets completely filled.
bool getOneC(struct dmasinfos *dmainfo, uint8_t *c)
{
	if (dmainfo->hdma->Instance->CNDTR != dmainfo->decr)
	{
		*c = ((uint8_t*) (dmainfo->start))[dmainfo->rd_pointer-dmainfo->start];
		dmainfo->rd_pointer++;
		if (dmainfo->rd_pointer>dmainfo->end)
		{
			dmainfo->rd_pointer = dmainfo->start;
			dmainfo->decr = SERIAL_BUFFER_RX_SIZE;
		}
		else
     dmainfo->decr--;

		return true;
	}
	else 
		return false;
}

// public
bool getOneCfromUSB(uint8_t *c)
{
	return getOneC(&serialdmaUSB,c);
}

bool getOneCfromPin(uint8_t *c)
{
	return getOneC(&serialdmaCMOS,c);
}

/*
void getChar(void)
{
	static int x = 3;
	int y = 5;
	volatile char c;
	
	if (USART2->SR&(1<<5))
	{
		c = USART2->DR;
		setPix(x,y,1);
		x++;
	}
}
*/

