/* Includes ------------------------------------------------------------------*/
#include "stm32l4xx_hal.h"
/* USER CODE BEGIN Includes */
#include "string.h"
#include "globals.h"
#include "serial.h"
#include "font8x8_latin.h"
#include "video.h"
#include "terminal.h"
/* USER CODE END Includes */


// private

struct terminalState
{
	int x,y;
} tState;

void initTerminal(void) // reverse chars
{
	int i,j,k;
	uint8_t c_in, c_out;
	
	for (i=0;i<128;i++)
	{
		for (j=0;j<8;j++)
		{
			c_out = 0;
		  c_in = font8x8_basic[i][j];
			for (k=0;k<8;k++)
			{
				if (c_in & (1<<k))
					c_out |= (1<<(7-k));
			}
			font8x8_basic[i][j] = c_out;
		}
	}
	for (i=0;i<96;i++)
	{
		for (j=0;j<8;j++)
		{
			c_out = 0;
		  c_in = font8x8_ext_latin[i][j];
			for (k=0;k<8;k++)
			{
				if (c_in & (1<<k))
					c_out |= (1<<(7-k));
			}
			font8x8_ext_latin[i][j] = c_out;
		}
	}	
}


void advanceCursor(void)
{
		tState.x++;
		if (tState.x>=NB_COLUMNS)
		{
			tState.x = 0;
			tState.y++;
			if (tState.y>=NB_LINES)
				tState.y = 0;
		}		
}

void moveCursorDown(void)
{
			tState.y++;
			if (tState.y>=NB_LINES)
				tState.y = 0;	
}

void printC(int x, int y, uint8_t c, uint8_t *p)
{
	uint8_t byte;
	uint8_t *start_address;
	int i;

	start_address = &fb[y<<3][x];
	
	for (i=0;i<8;i++)
	{
		byte = (uint8_t ) *p++;
		*(start_address) = byte;
		start_address+= VID_HSIZE+1;
	}
}

// public
// call only once
void resetTerminal(void)
{
	tState.x = 0;
	tState.y = 0;
//	reverseChars();
}


void processChar(uint8_t c)
{
	
	if ((c>=32)&&(c<128))
	{
		printC(tState.x,tState.y,c,(uint8_t*)&font8x8_basic[c][0]);
		advanceCursor();
		return;
	}
	if (c>=0xA0)
	{
		printC(tState.x,tState.y,c,(uint8_t*)&font8x8_ext_latin[c-0xA0][0]);
		advanceCursor();		
		return;
	}
	if (c<32)
	{
		switch(c)
		{
			case 9 : 
				printC(tState.x,tState.y,c,(uint8_t*)&font8x8_basic[' '][0]); advanceCursor();
				printC(tState.x,tState.y,c,(uint8_t*)&font8x8_basic[' '][0]); advanceCursor();
				printC(tState.x,tState.y,c,(uint8_t*)&font8x8_basic[' '][0]); advanceCursor();
				printC(tState.x,tState.y,c,(uint8_t*)&font8x8_basic[' '][0]); advanceCursor();
				break;
			case 0x0A: moveCursorDown(); break;
			case 0x0D: tState.x = 0; moveCursorDown(); break;
			case 12 : vidClearScreen(); tState.x = 0; tState.y = 0; break;
		}
	}

}

