/* Includes ------------------------------------------------------------------*/
#include "stm32l4xx_hal.h"
/* USER CODE BEGIN Includes */
#include "string.h"
#include "globals.h"
#include "video.h"

/* USER CODE END Includes */


//#define VTOTAL	52							/* Total bytes to send through SPI */
__align(4) uint8_t fb[VID_VSIZE][VID_HSIZE+1];	/* Frame buffer including h front borch*/
static volatile uint16_t vline;				/* The current line being drawn */
static volatile uint32_t vflag;				/* When 1, the SPI DMA request can draw on the screen */
static volatile uint32_t vdraw;				/* Used to increment vline every 3 drawn lines */ 
bool firstLine = true;

uint8_t *framep = &fb[0][0]; // won't change
uint8_t *curlinep = &fb[0][0]; // will change after every line sent by DMA

extern SPI_HandleTypeDef hspi1;
extern DMA_HandleTypeDef hdma_spi1_tx;
extern TIM_HandleTypeDef htim1;
extern TIM_HandleTypeDef htim2;



typedef struct
{
  __IO uint32_t ISR;   /*!< DMA interrupt status register */
  __IO uint32_t Reserved0;
  __IO uint32_t IFCR;  /*!< DMA interrupt flag clear register */
} DMA_Base_Registers;


void myHAL_DMA_Start_IT(DMA_HandleTypeDef *hdma, uint32_t SrcAddress, uint32_t DstAddress, uint32_t DataLength)
{
  /* calculate DMA base and stream number */
  DMA_Base_Registers *regs = (DMA_Base_Registers *)hdma->DmaBaseAddress;
     
  /* Clear all flags */
  hdma->DmaBaseAddress->IFCR = (DMA_ISR_GIF1 << (hdma->ChannelIndex & 0x1CU));
	
  /* Configure the source, destination address and the data length */ 

  /* Configure DMA Stream data length */
  hdma->Instance->CNDTR = DataLength;

  /* Memory to Peripheral */

	/* Configure DMA Stream destination address */
	hdma->Instance->CPAR = DstAddress;

	/* Configure DMA Stream source address */
	hdma->Instance->CMAR = SrcAddress;

	
	/* Clear all interrupt flags at correct offset within the register */
	regs->IFCR = 0x3FU << hdma->ChannelIndex;
	
	/* Enable Common interrupts*/
	hdma->Instance->CCR  |= DMA_IT_TC | DMA_IT_TE /*| DMA_IT_DME*/;
//	hdma->Instance->FCR |= DMA_IT_FE;
	
	/* Enable the Peripheral */
	__HAL_DMA_ENABLE(hdma);

}

void myHAL_DMA_Start_IT2(DMA_HandleTypeDef *hdma, uint32_t SrcAddress, uint32_t DstAddress, uint32_t DataLength)
{
  /* calculate DMA base and stream number */
  DMA_Base_Registers *regs = (DMA_Base_Registers *)hdma->DmaBaseAddress;
 
    /* Disable the peripheral */
    __HAL_DMA_DISABLE(hdma); // NEW
	
  /* Clear all flags */
  hdma->DmaBaseAddress->IFCR = (DMA_ISR_GIF1 << (hdma->ChannelIndex & 0x1CU)); // OK
	
  /* Configure the source, destination address and the data length */ 

  /* Configure DMA Stream data length */
  hdma->Instance->CNDTR = DataLength;

  /* Memory to Peripheral */

	/* Configure DMA Stream destination address */
	hdma->Instance->CPAR = DstAddress;

	/* Configure DMA Stream source address */
	hdma->Instance->CMAR = SrcAddress;

	
	/* Clear all interrupt flags at correct offset within the register */
	regs->IFCR = 0x3FU << hdma->ChannelIndex;
	
      __HAL_DMA_DISABLE_IT(hdma, DMA_IT_HT);	
	/* Enable Common interrupts*/
	 __HAL_DMA_ENABLE_IT(hdma, (DMA_IT_TC | DMA_IT_TE));

    /* Enable the Peripheral */
    __HAL_DMA_ENABLE(hdma);

}

void myHAL_SPI_Transmit_DMA(SPI_HandleTypeDef *hspi, uint8_t *pData, uint16_t Size)
{
  /* Configure communication direction : 1Line */
	SPI_1LINE_TX(hspi); // OK
	
	CLEAR_BIT(hspi->Instance->CR2, SPI_CR2_LDMATX); // NEW
	
  /* Enable the Tx DMA Stream/Channel */
  myHAL_DMA_Start_IT2(hspi->hdmatx, (uint32_t) pData, (uint32_t)&hspi->Instance->DR, Size);

  /* Check if the SPI is already enabled */
  if ((hspi->Instance->CR1 & SPI_CR1_SPE) != SPI_CR1_SPE) // OK
  {
    /* Enable SPI peripheral */
    __HAL_SPI_ENABLE(hspi);
  }

  /* Enable the SPI Error Interrupt Bit */
  __HAL_SPI_ENABLE_IT(hspi, (SPI_IT_ERR)); // OK

  /* Enable Tx DMA Request */
  SET_BIT(hspi->Instance->CR2, SPI_CR2_TXDMAEN);

}


//*****************************************************************************
//	This irq is generated at the end of the horizontal back porch.
//	Test if inside a valid vertical start frame (vflag variable), 
//	and start the DMA to output a single frame buffer line through the SPI device.
//*****************************************************************************
void HAL_TIM_OC_DelayElapsedCallback(TIM_HandleTypeDef *htim)
{
	static uint8_t firstcall = 1;
	
	if (htim==&htim1)
	{
		 if ((htim->Instance->CCMR1 & TIM_CCMR1_CC2S) == 0x00U) // CC2 TIM1 Output Compare Event
		 {
			 if (vflag) // is this a visible line ?
			 { // start sending data to VGA analog pin
				 if (firstcall)
				 {
					myHAL_SPI_Transmit_DMA(&hspi1,curlinep,VID_HSIZE_R); // far toooo slow
					firstcall = 0;
				 }
				 else
				 {
					SET_BIT(hspi1.Instance->CR2, SPI_CR2_TXDMAEN); // starts the DMA					 
				 }
			   firstLine = false; // do not remove, introduces a slight delay makes stability.
			 }
		 }
		 else
		 {
			 while(1) // in order to detect this problem. Should never happen.
			 {
			 }
		 }
	}

//*****************************************************************************
//	This irq is generated at the end of the vertical back porch.
//	Sets the 'vflag' variable to 1 (valid vertical frame).
//*****************************************************************************	
	else if (htim==&htim2)
	{
		 if ((htim->Instance->CCMR1 & TIM_CCMR1_CC2S) == 0x00U) // CC2 TIM2
		 {
				vflag = 1; // from now, we can display the lines
			  firstLine = true;
				HAL_PWR_EnableSleepOnExit();
		 }
	}
}

//*****************************************************************************
//	This interrupt is generated at the end of every line.
//	It will increment the line number, set the corresponding line pointer
//	in the DMA register and setup everything up in advance so that we will be able to
//  start the DMA very fast after next front porch.
//*****************************************************************************
/**
  * @brief This function handles DMA2 stream3 global interrupt.
*/

/**
  * @brief This function handles DMA1 channel3 global interrupt.
  */
void DMA1_Channel3_IRQHandler(void)
{
//	HAL_DMA_IRQHandler(&hdma_spi1_tx);
	SPI_HandleTypeDef *hspi = &hspi1;
//	goto suite;
//	calculate DMA base and stream number 
//  DMA_Base_Registers *regs = (DMA_Base_Registers *)hdma_spi1_tx.DmaBaseAddress;
  __HAL_DMA_DISABLE_IT(&hdma_spi1_tx, DMA_IT_TE | DMA_IT_TC); // OK
	
//	Clear the transfer complete flag 
	hdma_spi1_tx.DmaBaseAddress->IFCR = (DMA_ISR_TCIF1 << (hdma_spi1_tx.ChannelIndex & 0x1CU)); // OK

	/* Disable ERR interrupt */
	__HAL_SPI_DISABLE_IT(hspi, SPI_IT_ERR); // OK
	
//	Disable Tx DMA Request 
	CLEAR_BIT(hspi->Instance->CR2, SPI_CR2_TXDMAEN);  // OK
	
	__HAL_SPI_CLEAR_OVRFLAG(hspi); // OK

	
	vline++;

	if (vline == VID_VSIZE)
	{
		vline = vflag = 0;
		curlinep = framep;
		HAL_PWR_DisableSleepOnExit();
	} 
	else 
	{
		curlinep+=VID_HSIZE+1;
	}	
	
	// setup DMA for next time
	// this is the stripped version of HAL_SPI_Transmit_DMA
	// thus much faster


	CLEAR_BIT(hspi->Instance->CR2, SPI_CR2_LDMATX); // NEW
	
	__HAL_DMA_DISABLE(&hdma_spi1_tx);
	
  /* Clear all flags */
  hdma_spi1_tx.DmaBaseAddress->IFCR = (DMA_ISR_GIF1 << (hdma_spi1_tx.ChannelIndex & 0x1CU)); // OK
	
  /* Configure the source, destination address and the data length */ 


	hdma_spi1_tx.Instance->CNDTR = VID_HSIZE+1;
	hdma_spi1_tx.Instance->CMAR = (uint32_t) curlinep;
	hdma_spi1_tx.Instance->CCR  |= DMA_IT_TC | DMA_IT_TE /*| DMA_IT_DME*/;
	
 /* calculate DMA base and stream number */
  DMA_Base_Registers *regs = (DMA_Base_Registers *)hdma_spi1_tx.DmaBaseAddress;
	
	/* Clear all interrupt flags at correct offset within the register */
	regs->IFCR = 0x3FU << hdma_spi1_tx.ChannelIndex;
	
  __HAL_DMA_DISABLE_IT(&hdma_spi1_tx, DMA_IT_HT);	
	
  /* Check if the SPI is already enabled */
  if ((hspi->Instance->CR1 & SPI_CR1_SPE) != SPI_CR1_SPE) // OK
  {
    /* Enable SPI peripheral */
    __HAL_SPI_ENABLE(hspi);
  }
	
    /* Enable the Peripheral */
	__HAL_DMA_ENABLE(&hdma_spi1_tx);
	__HAL_SPI_ENABLE_IT(&hspi1, (SPI_IT_ERR));	
}


/*
void DMA2_Stream3_IRQHandler(void)
{
	SPI_HandleTypeDef *hspi = &hspi1;
//	calculate DMA base and stream number 
  DMA_Base_Registers *regs = (DMA_Base_Registers *)hdma_spi1_tx.StreamBaseAddress;
	
//	Clear the transfer complete flag 
	regs->IFCR = DMA_FLAG_TCIF0_4 << hdma_spi1_tx.StreamIndex;
//	Disable the transfer complete interrupt 
	hdma_spi1_tx.Instance->CR  &= ~(DMA_IT_TC);
	__HAL_SPI_DISABLE_IT(hspi, SPI_IT_ERR);
//	Disable Tx DMA Request 
	CLEAR_BIT(hspi->Instance->CR2, SPI_CR2_TXDMAEN);   
	__HAL_SPI_CLEAR_OVRFLAG(hspi);

	
	vline++;

	if (vline == VID_VSIZE)
	{
		vline = vflag = 0;
		curlinep = framep;
		HAL_PWR_DisableSleepOnExit();
	} 
	else 
	{
		curlinep+=VID_HSIZE+1;
	}	
	
	// setup DMA for next time
	// this is the stripped version of HAL_SPI_Transmit_DMA
	// thus much faster

	hdma_spi1_tx.Instance->NDTR = VID_HSIZE+1;
	hdma_spi1_tx.Instance->M0AR = (uint32_t) curlinep;
	hdma_spi1_tx.Instance->CR  |= DMA_IT_TC | DMA_IT_TE | DMA_IT_DME;

	__HAL_DMA_ENABLE(&hdma_spi1_tx);
	__HAL_SPI_ENABLE_IT(&hspi1, (SPI_IT_ERR));	
}
*/
// public

// val : 1 or 0
void setPix(uint16_t x, uint16_t y, int val)
{
	if (x>=VID_PIXELS_X)
		return;
	if (y>=VID_PIXELS_Y)
		return;
	
	if (val)
		fb[y][x>>3] |= 1 << (7-(x&0x7)); // the SPI sends the MSB first.
	else
		fb[y][x>>3] &= ~(1 << (7-(x&0x7))); 
}


void vidInit(void)
{
	vdraw = vline = vflag = 0;
	curlinep = framep;
}


void vidClearScreen(void)
{

//	memset(framep,0,VID_SIZE); // optim equiv to :
	
	int x, y;

	for (y = 0; y < VID_VSIZE; y++)
	{
		for (x = 0; x < VID_HSIZE+1; x++)
		{
			fb[y][x] = 0;
		}
	}
}

// fill in a start image
// draw a square around the screen
void vidDrawFrame(void)
{
	int x;
	int y;

	for (x=0;x<VID_PIXELS_X;x+=VID_PIXELS_X-1)
  {
		for (y=0;y<VID_PIXELS_Y;y++)
		{
			setPix(x,y,1);
		}
	}

	for (x=0;x<VID_PIXELS_X;x++)
  {
		for (y=0;y<VID_PIXELS_Y;y+=VID_PIXELS_Y-1)
		{
			setPix(x,y,1);
		}
	}	
}

// fill in a start image
// draws a grid around the screen
void vidDrawGrid(void)
{
	int x;
	int y;

	for (x=0;x<VID_PIXELS_X;x+=16)
  {
		for (y=0;y<VID_PIXELS_Y;y++)
		{
			setPix(x,y,1);
		}
	}

	for (x=0;x<VID_PIXELS_X;x++)
  {
		for (y=0;y<VID_PIXELS_Y;y+=16)
		{
			setPix(x,y,1);
		}
	}	
}



